@@ -1,26 +1,46 @@
 # -*- coding: utf-8 -*-
 """MOFChecker: Basic sanity checks for MOFs"""
+import logging
 import os
 import warnings
 from collections import OrderedDict
 from pathlib import Path
 from typing import List, Union
 
+import networkx as nx
 import numpy as np
+from networkx.algorithms.graph_hashing import weisfeiler_lehman_graph_hash
 from pymatgen import Structure
 from pymatgen.analysis.graphs import ConnectedSite, StructureGraph
-from pymatgen.analysis.local_env import CrystalNN, JmolNN, LocalStructOrderParams
+from pymatgen.analysis.local_env import (
+    BrunnerNN_relative,
+    CrystalNN,
+    EconNN,
+    JmolNN,
+    LocalStructOrderParams,
+    MinimumDistanceNN,
+)
 from pymatgen.io.cif import CifParser
 
 from ._version import get_versions
+from .checks.zeopp import check_if_porous
 from .definitions import CHECK_DESCRIPTIONS, EXPECTED_CHECK_VALUES, OP_DEF
 from .utils import (
     HighCoordinationNumber,
     LowCoordinationNumber,
     NoMetal,
     NoOpenDefined,
+    _check_if_ordered,
+    _check_metal_coordination,
+    _guess_underbound_nitrogen_cn2,
+    _guess_underbound_nitrogen_cn3,
+    _is_any_neighbor_metal,
+    _maximum_angle,
+    _vdw_radius_neighbors,
+    get_charges,
     get_overlaps,
     get_subgraphs_as_molecules_all,
+    is_metal,
 )
 
 __version__ = get_versions()["version"]
@@ -28,6 +48,38 @@ del get_versions
 
 __all__ = ["__version__", "MOFChecker"]
 
+MOFCheckLogger = logging.getLogger(__name__)
+MOFCheckLogger.setLevel(logging.DEBUG)
+
+try:
+    from openbabel import pybel  # pylint:disable=import-outside-toplevel, unused-import
+
+    HAS_OPENBABEL = True
+except ImportError:
+    warnings.warn(
+        "For the charge check openbabel needs to be installed. \
+    This can be done, for example using conda install openbabel"
+    )
+    HAS_OPENBABEL = False
+
+
+def construct_clean_graph(
+    structure: Structure, structure_graph: StructureGraph
+) -> nx.Graph:
+    """Creates a networkx graph with atom numbers as node labels"""
+    edges = {
+        (str(structure[u].specie), str(structure[v].specie))
+        for u, v, d in structure_graph.graph.edges(keys=False, data=True)
+    }
+    graph = nx.Graph()
+    graph.add_edges_from(edges)
+
+    for node in graph.nodes:
+
+        graph.nodes[node]["label"] = node
+
+    return graph
+
 
 class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-public-methods
     """MOFChecker performs basic sanity checks for MOFs"""
@@ -37,12 +89,19 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
 
         Args:
             structure (Structure): pymatgen Structure object
+
+        Raises:
+            NotImplementedError in the case of partial occupancies
         """
         self.structure = structure
+        _check_if_ordered(structure)
         self.metal_indices = [
-            i for i, species in enumerate(self.structure.species) if species.is_metal
+            i for i, site in enumerate(self.structure) if is_metal(site)
         ]
-        self.porous_adjustment = True
+
+        self.porous_adjustment = False
+        self.charges = None
+        self._porous = ""
         self.metal_features = None
         self._open_indices: set = set()
         self._has_oms = None
@@ -55,7 +114,9 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
             i for i, species in enumerate(self.structure.species) if str(species) == "C"
         ]
         self.h_indices = [
-            i for i, species in enumerate(self.structure.species) if str(species) == "H"
+            i
+            for i, species in enumerate(self.structure.species)
+            if str(species) in ["H", "D", "T"]
         ]
         self.n_indices = [
             i for i, species in enumerate(self.structure.species) if str(species) == "N"
@@ -69,8 +130,10 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
         self.check_descriptions = CHECK_DESCRIPTIONS
 
         self._graph = None
+        self._nx_graph = None
         self._connected_sites = {}
         self._cns = {}
+        self._set_cnn()
 
     def _set_filename(self, path):
         self._filename = os.path.abspath(path)
@@ -87,6 +150,23 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
         """Return the indices of overlapping atoms"""
         return self._get_atomic_overlaps()
 
+    @property
+    def graph_hash(self):
+        """Return the Weisfeiler-Lehman graph hash.
+        Hashes are identical for isomorphic graphs
+        (taking the atomic kinds into account)
+        and there are guarantees that non-isomorphic graphs will get different hashes.
+        """
+        return weisfeiler_lehman_graph_hash(self.nx_graph, node_attr="label")
+
+    @property
+    def scaffold_hash(self):
+        """Return the Weisfeiler-Lehman graph hash.
+        Hashes are identical for isomorphic graphs and there are
+        guarantees that non-isomorphic graphs will get different hashes.
+        """
+        return weisfeiler_lehman_graph_hash(self.nx_graph)
+
     @property
     def has_atomic_overlaps(self):
         """Check if there are any overlaps in the structure"""
@@ -177,8 +257,9 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
         for site_index in self.c_indices:
             cn = self.get_cn(site_index)  # pylint:disable=invalid-name
             if cn > 4:
-                overvalent_c = True
-                break
+                if not _is_any_neighbor_metal(self.get_connected_sites(site_index)):
+                    overvalent_c = True
+                    break
         self._overvalent_c = overvalent_c
 
     def _has_overvalent_h(self):
@@ -186,11 +267,22 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
         for site_index in self.h_indices:
             cn = self.get_cn(site_index)  # pylint:disable=invalid-name
             if cn > 1:
-                overvalent_h = True
-                break
+                if not _is_any_neighbor_metal(self.get_connected_sites(site_index)):
+                    overvalent_h = True
+                    break
         self._overvalent_h = overvalent_h
 
-    def _has_undercoordinated_carbon(self, tolerance=10):
+    def _has_overvalent_n(self):
+        overvalent_n = False
+        for site_index in self.n_indices:
+            cn = self.get_cn(site_index)  # pylint:disable=invalid-name
+            if cn > 4:
+                if not _is_any_neighbor_metal(self.get_connected_sites(site_index)):
+                    overvalent_n = True
+                    break
+        self._overvalent_n = overvalent_n
+
+    def _has_undercoordinated_carbon(self, tolerance: int = 10):
         """Idea is that carbon should at least have three neighbors if it is not sp1.
         In sp1 case it is linear. So we can just check if there are carbons with
         non-linear coordination with less than three neighbors. An example in CoRE
@@ -202,48 +294,104 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
         for site_index in self.c_indices:
             cn = self.get_cn(site_index)  # pylint:disable=invalid-name
             if cn == 2:
-                # ToDo: Check if it is bound to metal, then it might be a carbide
                 neighbors = self.get_connected_sites(site_index)
-                angle = self.structure.get_angle(
-                    site_index, neighbors[0].index, neighbors[1].index
+                angle = _maximum_angle(
+                    self.structure.get_angle(
+                        site_index, neighbors[0].index, neighbors[1].index
+                    )
                 )
-                if np.abs(90 - angle) > tolerance:
-                    undercoordinated_carbon = True
-                    break
+                if (np.abs(180 - angle) > tolerance) or (np.abs(180 - 0) > tolerance):
+                    if (not is_metal(neighbors[0].site)) or (
+                        not is_metal(neighbors[1].site)
+                    ):
+                        if len(_vdw_radius_neighbors(self.structure, site_index)) <= 2:
+                            undercoordinated_carbon = True
+                            break
         self._undercoordinated_carbon = undercoordinated_carbon
 
-    def _has_undercoordinated_nitrogen(self, tolerance=10):
+    def _has_undercoordinated_nitrogen(self, tolerance: int = 15):
         """
-        Captures missing hydrogens on amino groups.
-        Basically two common cases:
-            1. Have the N on a carbon and no hydrogen at all
-            2. (not that common) due to incorrect symmetry resolution
-                we have only one h in a bent orientation
+        Attempts to captures missing hydrogens on nitrogen groups
+        using heuristics
         """
         undercoordinated_nitrogen = False
         for site_index in self.n_indices:
             cn = self.get_cn(site_index)  # pylint:disable=invalid-name
+            neighbors = self.get_connected_sites(site_index)
             if cn == 1:
                 # this is suspicous, but it also might a CN which is perfectly fine.
                 # to check this, we first see if the neighbor is carbon
-                # and then what its coordination number is
-                neighbors = self.get_connected_sites(site_index)
-                if (self.get_cn(neighbors[0].index) > 2) and (
-                    str(neighbors[0].periodic_site.specie) == "C"
-                ):
-                    undercoordinated_nitrogen = True
-                    break
-            if cn == 2:
-                # ToDo: Check if it is bound to metal, then it might be a nitride
-                neighbors = self.get_connected_sites(site_index)
-                angle = self.structure.get_angle(
-                    site_index, neighbors[0].index, neighbors[1].index
+                # and then what its coordination number is. If it is greater than 2
+                # then we likely do not have a CN for which the carbon should be a
+                # linear sp one
+                print(
+                    site_index,
+                    cn,
+                    neighbors,
+                    self.get_cn(neighbors[0].index),
+                    self.get_connected_sites(neighbors[0].index),
                 )
-                if np.abs(90 - angle) > tolerance:
+                if (self.get_cn(neighbors[0].index) > 2) and not neighbors[
+                    0
+                ].site.specie.is_metal:
                     undercoordinated_nitrogen = True
                     break
+            elif cn == 2:
+                undercoordinated_nitrogen = _guess_underbound_nitrogen_cn2(
+                    self.structure,
+                    site_index,
+                    neighbors,
+                    self.get_connected_sites(neighbors[0].index),
+                    self.get_connected_sites(neighbors[1].index),
+                    tolerance,
+                )
+                if undercoordinated_nitrogen:
+                    break
+            elif cn == 3:
+                undercoordinated_nitrogen = _guess_underbound_nitrogen_cn3(
+                    self.structure, site_index, neighbors, tolerance
+                )
+                if undercoordinated_nitrogen:
+                    break
+
         self._undercoordinated_nitrogen = undercoordinated_nitrogen
 
+    def _has_high_charges(self, threshold=3) -> Union[bool, None]:
+        if (self.charges is None) and HAS_OPENBABEL:
+            self.charges = get_charges(self.structure)
+
+        if isinstance(self.charges, list):
+            if np.sum(np.abs(self.charges) > threshold):
+                return True
+        else:
+            return None
+
+        return False
+
+    def _is_porous(self) -> Union[bool, None]:
+        if self._porous == "":
+            self._porous = check_if_porous(self.structure)
+        return self._porous
+
+    @property
+    def is_porous(self) -> Union[bool, None]:
+        """Returns True if the MOF is porous according to the CoRE-MOF definition.
+        Returns None if the check could not be run successfully."""
+        return self._is_porous()
+
+    @property
+    def has_high_charges(self) -> Union[bool, None]:
+        """Check if the structure has unreasonably high EqEq charges.
+        Returns None if the check could not be run successfully."""
+        return self._has_high_charges()
+
+    @property
+    def nx_graph(self) -> nx.Graph:
+        """Returns a networkx graph with atom numbers as node labels"""
+        if self._nx_graph is None:
+            _ = self.graph
+        return self._nx_graph
+
     @property
     def graph(self) -> StructureGraph:
         """pymatgen structure graph."""
@@ -251,6 +399,7 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
             self._graph = StructureGraph.with_local_env_strategy(
                 self.structure, self._cnn
             )
+            self._nx_graph = construct_clean_graph(self.structure, self._graph)
         return self._graph
 
     def get_connected_sites(self, site_index) -> List[ConnectedSite]:
@@ -278,7 +427,6 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
         if site_index not in self._cns:
             with warnings.catch_warnings():
                 warnings.simplefilter("ignore")
-                self._set_cnn()
                 self._cns[site_index] = self._cnn.get_cn(self.structure, site_index)
         return self._cns[site_index]
 
@@ -306,23 +454,21 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
         """Returns true if there is a isolated floating atom or molecule"""
         return self._has_stray_molecules()
 
-    def _has_lone_atom(self) -> bool:
-        self._set_cnn()
-        for site in range(len(self.structure)):
-            nbr = self.get_connected_sites(site)
+    def _has_lone_atom(self, safe: bool = True) -> bool:
+        for site_index in range(len(self.structure)):
+            nbr = self.get_connected_sites(site_index)
             if not nbr:
-                return True
+                lone = True
+                # safe option checks if there is really nothing around 1.5 * VdW radius
+                # this is useful as sometimes the NN method is off, one example for this
+                # is FEZTIP
+                if safe:
+                    if _vdw_radius_neighbors(self.structure, site_index):
+                        lone = False
+                if lone:
+                    return True
         return False
 
-    def _has_overvalent_n(self):
-        overvalent_n = False
-        for site_index in self.n_indices:
-            cn = self.get_cn(site_index)  # pylint:disable=invalid-name
-            if cn > 4:
-                overvalent_n = True
-                break
-        self._overvalent_n = overvalent_n
-
     @classmethod
     def _from_file(cls, path: str):
         structure = Structure.from_file(path)
@@ -348,7 +494,7 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
             omscls._set_filename(path)  # pylint:disable=protected-access
             return omscls
 
-    def _set_cnn(self, method="JmolNN", porous_adjustment: bool = True):
+    def _set_cnn(self, method="jmolnn", porous_adjustment: bool = False):
         if self._cnn_method == method.lower():
             return
         self._cnn_method = method.lower()
@@ -356,6 +502,12 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
         self.porous_adjustment = porous_adjustment
         if method.lower() == "crystalnn":
             self._cnn = CrystalNN(porous_adjustment=self.porous_adjustment)
+        elif method.lower() == "econnn":
+            self._cnn = EconNN()
+        elif method.lower() == "brunnernn":
+            self._cnn = BrunnerNN_relative()
+        elif method.lower() == "minimumdistance":
+            self._cnn = MinimumDistanceNN()
         else:
             self._cnn = JmolNN()
 
@@ -394,7 +546,7 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
 
     def is_site_open(self, site_index: int) -> bool:
         """Check for a site if is open (based on the values of
-        some coordination geomeetry fingerprints)
+        some coordination geometry fingerprints)
 
         Args:
             site_index (int): Index of the site in the structure
@@ -405,7 +557,6 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
         if site_index not in self._open_indices:
             try:
                 _, _names, lsop, is_open, weights = self._get_ops_for_site(site_index)
-                # print(list(zip(names, lsop)))
                 site_open = MOFChecker._check_if_open(lsop, is_open, weights)
                 if site_open:
                     self._open_indices.add(site_index)
@@ -456,7 +607,6 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
         return descriptors
 
     def _has_stray_molecules(self) -> bool:
-        self._set_cnn()
         molecules = get_subgraphs_as_molecules_all(self.graph)
         if len(molecules) > 0:
             return True
@@ -472,6 +622,8 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
         result_dict = OrderedDict(
             (
                 ("name", self.name),
+                ("graph_hash", self.graph_hash),
+                ("formula", self.formula),
                 ("path", self._filename),
                 ("density", self.density),
                 ("has_oms", self.has_oms),
@@ -486,6 +638,9 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
                 ("has_metal", self.has_metal),
                 ("has_lone_atom", self.has_lone_atom),
                 ("has_lone_molecule", self.has_lone_molecule),
+                ("has_high_charges", self.has_high_charges),
+                ("has_undercoordinated_metal", self.has_undercoordinated_metal),
+                ("is_porous", self.is_porous),
             )
         )
         return result_dict
@@ -507,6 +662,20 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
 
         return descriptordict
 
+    def _has_low_metal_coordination(self):
+        for site_index in self.metal_indices:
+            if _check_metal_coordination(
+                self.structure[site_index], self.get_cn(site_index)
+            ):
+                if self.is_site_open(site_index):
+                    return True
+        return False
+
+    @property
+    def has_undercoordinated_metal(self):
+        """Check if a metal has unusually low coordination"""
+        return self._has_low_metal_coordination()
+
     def get_metal_descriptors(self) -> dict:
         """Return local structure order parameters for coordination number (CN),
         element string and wheter site is open or not. Key is the site index.
@@ -532,9 +701,9 @@ class MOFChecker:  # pylint:disable=too-many-instance-attributes, too-many-publi
     def has_oms(self) -> bool:
         """True if the structure contains open metal sites (OMS).
         Also returns True in case of low coordination numbers (CN <=3)
-        which typically also means open coordination for MOFs.
-        For high coordination numbers, for which we do not have a good order
-        parameter for open structures. For this reason we return None even though
+        which typically indicate open coordination for MOFs.
+        For high coordination numbers, no good order parameter for open
+        structures is available, and so we return `None` even though
         this might change in a future release.
 
         Raises:
